import os
import pickle
import time
import numpy as np
import multiprocessing

this_dir, this_filename = os.path.split(__file__)

directory = 'binlens_images2'
directory = os.path.join(this_dir, directory)

import binlen

debug = False
def clean():
    files = os.listdir(directory)
    for f in files:
        if f.startswith('_tmpPMOIRED') and f.endswith('.pickle'):
            os.system('rm -rf '+os.path.join(directory, f))
        if f in [txt]:
            os.system('rm '+os.path.join(directory, f))
    return

def computeSparseParam(MJD, param, parallel=False):
    images = computeImages(MJD, param, parallel=parallel)
    fluxes = []
    mjd = []
    for k in images:
        mjd.append(k)
        if type(images[k])==str:
            with open(images[k], 'rb') as f:
                tmp = pickle.load(f)
            fluxes.append(np.sum(tmp['I']))
        else:
            fluxes.append(np.sum(images[k]['I']))

    fluxes = np.array(fluxes)
    fluxes = fluxes[np.argsort(mjd)]
    mjd = sorted(mjd)
    s = 'np.interp($MJD, %s, %s)'%('['+','.join(['%.2f'%x for x in mjd])+']',
                                   '['+','.join(['%.3f'%x for x in fluxes])+']')
    if False:
        print('images = {',end='')
        for k in images:
            print(str(k)+':', end='')
            if type(images[k])==str:
                print('"'+images[k]+'"', end=',')
            else:
                print(','.join(['"'+x+'":len%d'%len(images[k][x]) for x in images[k]]), end=',')
            print()
        print('}')

    return {'sparse':images, 'spectrum(mjd)':s}

_lastParam = []
_lastMJD = {}
# files = list(filter(lambda x: x.startswith('_tmp') and
#                         x.endswith('.pickle'), os.listdir()))
# for f in files:
#     os.remove(f)
def computeImages(MJD, param, parallel=False, debug=False):
    """
    parallel is actually much slower!
    """
    global _lastParam, _lastMJD

    useFiles=True # use files to store sparse images -> SLOW!
    useFiles=False #

    if parallel:
        preptime = 0
        t0 = time.time()
        nproc = min(multiprocessing.cpu_count(), len(MJD))

        Pool = multiprocessing.Pool(nproc)
        res = []
        for i,mjd in enumerate(MJD):
            res.append(Pool.apply_async(computeImages, ([mjd], param, ),
                {'parallel':False, 'debug':False}))
        Pool.close()
        Pool.join()
        tmp = [r.get(timeout=1) for r in res]
        res = {}
        for t in tmp:
            res.update(t)
        return res

    t0 = time.time()
    cols = ['t0', 'u0', 'tE', 'b', 'q', 'theta', 'rhos']
    if not 't0' in param and 'mjd0' in param:
        param['t0'] = param['mjd0']- 49999.5 

    _p = [param[c] for c in cols]
    if _p!=_lastParam:
        # -- reset model computing, remove old images
        #files = list(filter(lambda x: x.startswith('img.') and
        #                        x.endswith('.pickle') and
        #                        len(x)==18, os.listdir()))
        files = list(filter(lambda x: os.path.exists(x),
                            [_lastMJD[k] for k in _lastMJD if type(k)==str]))
        for f in files:
            os.remove(f)
        #os.system('rm img.????.??.pickle')
        if debug:
            print('D> parameters have changed! removing %d old image(s)'%len(files))
        # -- reset recent parameters and MJDs
        _lastParam = _p
        _lastMJD = {}

    res = {}
    if debug:
        print('D> -- computeImage starts')
    for mjd in MJD:
        if mjd in _lastMJD:
            if debug:
                print('D> same parameters and known MJD: re-using image')
            res[round(mjd, 2)] = _lastMJD[mjd]
            continue

        if debug:
            print('D> computing for epoch MJD=%f:'%mjd)
        t1 = time.time()
        tmp = [param[k] for k in cols]
        tmp += [round(mjd - 49999.5, 2)]

        if debug:
            print('D> computing:', tmp)

        # -- call Fortran function
        ng, xg, yg = binlen.binlen_image(*tmp)
        xg = xg[:ng]
        yg = yg[:ng]
        # -- from Fortran code
        if param['b']<=1:
            offx = param['b']*(-0.5+1/(1+param['q']))
        else:
            offx = param['b']/2 -param['q']/(1+param['q'])/param['b']
        xg += offx

        t1 = time.time()
        if useFiles:
            imgp = '_tmpPMOIRED%d.pickle'%np.random.randint(1e16)
            while os.path.exists(imgp):
                imgp = '_tmpPMOIRED%d.pickle'%np.random.randint(1e16)
            if debug:
                print('D> image computed in %.2fs'%(time.time()-t1))
            imageTxt2Sparse( X=xg, Y=yg,rhos=param['rhos'], debug=debug,
                            savefile=True, filename=imgp)
            _lastMJD[mjd] = imgp
        else:
            _lastMJD[mjd] = imageTxt2Sparse(X=xg, Y=yg, rhos=param['rhos'],
                                    debug=debug, savefile=False)
        if debug:
            print('D> conversion in %.2fs'%(time.time()-t1))
        res[round(mjd, 2)] = _lastMJD[mjd]

    if debug:
        print('D> -- computeImage done in %.2fs'%(time.time()-t0))
    #print("computeImages:", res.keys())

    return res

def lightCurve(MJD, param, stepping=None):
    """
    param: dictionnary containing ['t0', 'u0', 'tE', 'b', 'q', 'theta', 'rhos']

    MJD: list of MJDs

    stepping: use a interpolated light curve instead of for every date
    """
    cols = ['t0', 'u0', 'tE', 'b', 'q', 'theta', 'rhos']
    if not 't0' in param and 'mjd0' in param:
        param['t0'] = param['mjd0']- 49999.5 
    res = []
    if stepping is None:
        X = MJD
    else:
        X = np.linspace(np.min(MJD), np.max(MJD), int(np.ptp(MJD)/stepping)+1)

    for mjd in X:
        tmp = [param[k] for k in cols]
        tmp += [mjd - 49999.5]
        ng, xg, yg = binlen.binlen_image(*tmp)
        # -- normalise fluxes to flux of source
        # -- "0.002" -> grid size in fortran code
        fscale = (0.002/param['rhos'])**2/np.pi
        res.append(ng*fscale)
        #xg = xg[:ng]
        #yg = yg[:ng]
        # -- from Fortran code
        #if param['b']<=1:
        #    offx = param['b']*(-0.5+1/(1+param['q']))
        #else:
        #    offx = param['b']/2 -param['q']/(1+param['q'])/param['b']
        #xg += offx
    res = np.array(res)
    if not stepping is None:
        res = np.interp(MJD, X, res)
    return res

def imageTxt2Sparse(X=None, Y=None, rhos=None, filename=None, debug=debug,savefile=True):
    """
    this is  slow!
    """
    tinit = time.time()
    if X is None and Y is None:
        data = []
        with open(filename) as f:
            for l in f.readlines():
                data.append((np.double(l.split()[0]), np.double(l.split()[1])))
        X, Y = np.array([d[0] for d in data]), np.array([d[1] for d in data])
    if debug:
        print('D> X,Y have length', len(X))
    t = time.time()
    x0 = np.mean(X)
    y0 = np.mean(Y)
    if debug:
        print('D> X,Y have length', len(X))
        print('D> X: %.3f..%.3f'%(min(X), max(X)))
        print('D> Y: %.3f..%.3f'%(min(Y), max(Y)))

    #dx = 0.005 # original file
    dx = 0.005*4 # binned

    nx, ny = int(X.ptp()/dx)+1, int(Y.ptp()/dx)+1

    x = np.linspace(X.min(), X.max(), nx)
    y = np.linspace(Y.min(), Y.max(), ny)

    _X, _Y = np.meshgrid(x, y)
    _I = np.zeros(_X.shape)
    t = time.time()

    _i = np.int_(nx/(nx+1)*(Y-y.min())/dx)
    _j = np.int_(nx/(nx+1)*(X-x.min())/dx)

    # -- weirdly enough, the np call is slower than a loop!
    #np.add.at(_I, list(zip(_i, _j)), 1)

    for k in range(len(_i)):
        _I[_i[k], _j[k]] += 1

    if debug:
        print('D> resampling in %.2fs'%(time.time()-t))
    # -- generate sparse image
    sparse = {'x':[], 'y':[], 'I':[]}
    for i in range(ny):
        for j in range(nx):
            if _I[i,j]>0:
                sparse['x'].append(_X[i,j])
                sparse['y'].append(_Y[i,j])
                sparse['I'].append(_I[i,j])
    for k in sparse:
        sparse[k] = np.array(sparse[k])

    if not rhos is None:
        # -- normalise fluxes to flux of source
        # -- not sure how to justify the "0.002"! -> Maybe the stepping in ray tracing?
        # "0.002" -> grid size in fortran code
        fscale = (0.002/rhos)**2/np.pi
        sparse['I'] *= fscale
    if savefile:
        if not filename.endswith('.pickle'):
            filename += '.pickle'
        if debug:
            print('D> saving in', filename)
        with open(filename, 'wb') as f:
            pickle.dump(sparse, f)
        return filename
    else:
        return sparse



# /Library/Frameworks/Python.framework/Versions/3.8/bin/python3.8 -m cProfile -o out.txt speedtest.py
# > import pstats
# > p = pstats.Stats('out.txt')
# > p.sort_stats('time').print_stats(20)

import oifits, oimodels, os

directory = '/Users/amerand/DATA/Science/FSCMa/'
files = os.listdir(directory)
files = list(filter(lambda x: x.startswith('PIO'), files))

data = [oifits.loadOI(os.path.join(directory, f)) for f in files]

for d in data:
    # -- observable to fit
    d['fit'] = {'obs':['V2','T3PHI'], 
               'min error':{'V2':0.03, 'T3PHI':1.},
               'max error':{'V2':0.06, 'T3PHI':10.0},
               'mult error':{'V2':3., 'T3PHI':1.},
               }

param = {'disk,amp1':     -0.978, # +/- 0.075
         'disk,amp2':     -0.14, # +/- 0.14
         'disk,amp3':     0.074, # +/- 0.026
         'disk,diam':     38.03, # +/- 0.67
         'disk,f0':       0.804, # +/- 0.065
         'disk,f1':       1.31, # +/- 0.17
         'disk,i':        61.83, # +/- 0.95
         'disk,pa':       73.09, # +/- 0.81
         'disk,pa1':      84.19, # +/- 0.89
         'disk,pa2':      15.2, # +/- 17.6
         'disk,pa3':      -20.05, # +/- 6.30
         'disk,profile':  'doughnut' ,
         'disk,thick':    0.838, # +/- 0.068
         'inner,amp1':    -0.67, # +/- 0.24
         'inner,amp2':    -0.65, # +/- 0.77
         'inner,diam':    22.76, # +/- 1.77
         'inner,f0':      0.142, # +/- 0.063
         'inner,f1':      'inner,f0 * disk,f1 / disk,f0',
         'inner,i':       'disk,i' ,
         'inner,pa':      'disk,pa' ,
         'inner,pa1':     85.03, # +/- 6.07
         'inner,pa2':     20.1, # +/- 19.0
         'inner,profile': 'doughnut' ,
         'inner,thick':   0.97, # +/- 0.24
         'star,f0':       1.0 ,
         'star,f1':       -1 ,
         'star,ud':       0.20, # +/- 0.45
        }

res = oimodels.residualsOI(data, param, timeit=False)



#  correct tellurics for GRAVITY spectra
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
import scipy.interpolate
import os
import time
import pickle

try:
   # Necessary while Python versions below 3.9 are supported.
   import importlib_resources as resources
except ImportError:
   from importlib import resources

import pmoired
import pmoired.dpfit as dpfit


this_dir, this_filename = os.path.split(__file__)   # TODO clean this up
bin_file = 'transir_gravity.pckl'

try:
    # -- loading transnir
    with resources.as_file(resources.files('pmoired').joinpath(bin_file)) as p:
        with open(p, 'rb') as f:
            tran20, tran80, lbda = pickle.load(f)
except:
    print('creating transmission binary file')
    wv1 = '020' # 2mm Water Vapor
    f = fits.open(os.path.join(this_dir, 'transnir'+wv1+'.fits'))
    lbda = f[1].data['WAVELENGTH'][0].copy()
    tran = f[1].data['TRANSMISSION'][0].copy()
    f.close()
    tran = tran[(lbda<2.5)*(lbda>1.95)]
    lbda = lbda[(lbda<2.5)*(lbda>1.95)]
    tran20, lbda = 1.0*tran[::50], lbda[::50]
    f = fits.open(os.path.join(this_dir, 'emisnir'+wv1+'.fits'))
    lbda0 = f[1].data['WAVELENGTH'][0].copy()
    tran0 = f[1].data['EMISSION'][0].copy()
    tran20 += np.interp(lbda, lbda0[::50], tran0[::50])
    f.close()

    wv2 = '080' # 8mm Water Vapor
    f = fits.open(os.path.join(this_dir, 'transnir'+wv2+'.fits'))
    lbda = f[1].data['WAVELENGTH'][0].copy()
    tran = f[1].data['TRANSMISSION'][0].copy()
    f.close()
    tran = tran[(lbda<2.5)*(lbda>1.95)]
    lbda = lbda[(lbda<2.5)*(lbda>1.95)]
    tran80, lbda = 1.0*tran[::50], lbda[::50]
    f = fits.open(os.path.join(this_dir, 'emisnir'+wv2+'.fits'))
    lbda0 = f[1].data['WAVELENGTH'][0].copy()
    tran0 = f[1].data['EMISSION'][0].copy()
    tran80 += np.interp(lbda, lbda0[::50], tran0[::50])
    f.close()
    with open(os.path.join(this_dir, bin_file), 'wb') as f:
        pickle.dump((tran20, tran80, lbda), f)


def Ftran(l, param, retWL=False, retS=False):
    """
    'dl0', 'wl0', 'dl1',

    wl -> (wl-dl0)
    wl -> dli*(wl-wl0)**i + wl0
    """

    tmpL = 1.0*lbda

    # -- wavelength correction
    if 'dl0' in param.keys():
        tmpL -= param['dl0']
    if 'dl1' in param.keys() and 'wl0' in param.keys():
        tmpL = param['dl1']*(tmpL-param['wl0']) + param['wl0']

    for i in [2,3,4,5,6,7,8,9,10]:
        if 'dl'+str(i) in param.keys() and 'wl0' in param.keys():
            tmpL += param['dl'+str(i)]*(tmpL-param['wl0'])**i

    if retWL:
        return np.interp(l, tmpL, lbda)

    if 'pwv' in param.keys():
        tmpT = tran20 + (param['pwv']-2.0)*(tran80-tran20)/6.0
    else:
        tmpT = tran80

    if 'pow' in param.keys():
        tmpT = 1-np.abs(1-tmpT)**param['pow']

    # -- local corrections
    pc = filter(lambda x: x.startswith('p_'), param.keys())
    for x in pc:
        #
        wl0 = float(x.split('_')[1])
        if 'kern' in param:
            dwl = 1.5*param['kern']
        elif 'kern_min' in param and 'kern_max' in param:
            dwl = 0.75*(param['kern_min']+param['kern_max'])
        w = np.abs(tmpL-wl0)<dwl
        tmpT[w] = 1-(1-tmpT[w])**param[x]

    if 'kern' in param.keys():
        if 'kernp' in param.keys():
            kp = param['kernp']
        else:
            kp = 2.0
        kern = np.exp(-np.abs((tmpL-np.mean(tmpL))/param['kern'])**kp)
        kern /= np.sum(kern)
        tmpT = np.convolve(tmpT, kern, 'same')

    if 'kern_min' in param.keys() and 'kern_max' in param.keys():
        if 'kernp_min' in param.keys():
            kp_min = param['kernp_min']
        else:
            kp_min = 2.0
        if 'kernp_max' in param.keys():
            kp_max = param['kernp_max']
        else:
            kp_max = 2.0
        kern_min = np.exp(-np.abs((tmpL-np.mean(tmpL))/param['kern_min'])**kp_min)
        kern_max = np.exp(-np.abs((tmpL-np.mean(tmpL))/param['kern_max'])**kp_max)
        kern_min /= np.sum(kern_min)
        kern_max /= np.sum(kern_max)
        c = np.minimum(np.maximum((tmpL-np.min(l))/(np.ptp(l)), 0), 1)
        tmpT = (1-c)*np.convolve(tmpT, kern_min, 'same') + c*np.convolve(tmpT, kern_max, 'same')

    # -- spline shape
    X = sorted(filter(lambda x: x.startswith('xS'), param.keys()))
    Y = sorted(filter(lambda x: x.startswith('yS'), param.keys()))
    if len(X)==len(Y) and len(X):
        tmpS = scipy.interpolate.interp1d([param[x] for x in X], [param[y] for y in Y],
                                        kind='cubic', fill_value='extrapolate')(tmpL)
        tmpT *= tmpS
        if retS:
            return np.interp(l, tmpL, tmpS)
    return np.interp(l, tmpL, tmpT)

def removeTellurics(filename):
    f = fits.open(filename)
    if 'TELLURICS' in f:
        f.pop('TELLURICS')
        print('tellurics extension removed')
        f.writeto(filename, overwrite=True)
    else:
        print('tellurics not found, nothing to remove')
    f.close()
    return

def gravity(filename, quiet=True, save=True, wlmin=None, wlmax=None, avoid=None,
            fig=None, force=False, MR=True):
    """
    avoid: list of tuple of wlmin, wlmax to avoid in the fit (known) line
    """
    # -- LOAD DATA -------------------------------------------------
    f = fits.open(filename)

    # -- check if telluric already computed
    alreadyComputed = False
    for h in f:
        if 'EXTNAME' in h.header and h.header['EXTNAME']=='TELLURICS':
            alreadyComputed = True
    if not force and alreadyComputed:
        print('tellurics already computed for %s'%filename, end=' ')
        print('use "force=True" to recompute')
        f.close()
        return
    if 'ESO INS SPEC RES' in f[0].header:
        if 'MED' in f[0].header['ESO INS SPEC RES']:
            MR = True
        elif 'HIGH' in f[0].header['ESO INS SPEC RES']:
            MR = False
        else:
            print('Nothing to do for resolution', f[0].header['ESO INS SPEC RES'])
            return

    # -- init plot ----------------------------------------
    if not fig is None:
        quiet = False
    if not quiet:
        if fig is None:
            fig = 1
        plt.close(fig)
        plt.figure(fig, figsize=(10,5))
        plt.clf()
        plt.subplots_adjust(right=0.99, left=0.05)

    # -- HARDWIRED, DANGEROUS!!! -> works for POLA and JOINED
    try:
        wl = f[4].data['EFF_WAVE']*1e6
        if len(wl)<10:
            wl = f[3].data['EFF_WAVE']*1e6
    except:
        wl = f['OI_WAVELENGTH'].data['EFF_WAVE']*1e6

    if not quiet:
        print('WL (all):', wl.shape)

    # -- fitting domain:
    if wlmin is None:
        wlmin = 1.9
    if wlmax is None:
        wlmax = 2.5

    try:
        pola = f[0].header['HIERARCH ESO FT POLA MODE'].strip()=='SPLIT'
    except:
        pola = False

    # -- sum of fluxes, remove outliers -----------------------------
    sp, n = 0.0, 3
    fl = None
    for i in range(4):
        if pola:
            _sp = f[18].data['FLUX'][i,:]+f[22].data['FLUX'][i,:]
            _fl = np.logical_or(f[18].data['FLAG'][i,:], f[22].data['FLAG'][i,:])
        else:
            _sp = f[12].data['FLUX'][i,:].copy()
            _fl = f[12].data['FLAG'][i,:].copy()
        sp += _sp
        if fl is None:
            fl = _fl
        else:
            fl = np.logical_or(fl, _fl)
    # -- close FITS file
    f.close()

    # -- avoid some known lines
    w = (wl>wlmin)*(wl<wlmax)*~fl*~np.isnan(sp)*(sp!=0)

    if avoid is None:
        c = 3 if MR else 1
        # -- 2.058: HeI
        # -- 2.167: Br Gamma
        w = (np.abs(wl-2.058)>c*0.002)*(np.abs(wl-2.1665)>c*0.002)
    else:
        for _wlmin, _wlmax in avoid:
            w *= (wl<=_wlmin)+(wl>=_wlmax)

    if not quiet:
        print('WL (fit):', wl[w].shape)

    # -- FIT TELLURIC MODEL -------------------------------------------
    if MR:
        p = {'dl0':0.000, 'wl0':2.22, 'dl1':1.0, 'dl2':0.0, 'dl3':0.0,
             #'kern':1e-3,
             'kern_min':3e-3, 'kern_max':5e-3,
             #'kernp':1.0,
             'kernp_min':0.8, 'kernp_max':1.0,
             'pwv':2.0, 'pow':1.0, #'p_2.3717':0.8
             }
    else:
        p = {'dl0':0.0, 'wl0':2., 'dl1':1.0, 'dl2':0.0, 'dl3':0.0,
             'kern_min':2.8e-4, 'kern_max':3.4e-4, 'kernp_min':1.7, 'kernp_max':1.7,
             'pwv':2.0, 'pow':1,
             #'p_2.09913':0.8, 'p_2.11905':1.0, 'p_2.12825':1.0, 'p_2.1691':1.0,
             }

    # -- spectrum model using spline nodes
    Nn =  12 if MR else 35

    for i,x in enumerate(np.linspace(wl[w].min(), wl[w].max(), Nn)):
        test = True
        # -- makes sure the node is not in an avoidance zone:
        if not avoid is None:
            for _wlmin, _wlmax in avoid:
                if x>=_wlmin and x<=_wlmax:
                    test = False
        if test:
            p['xS'+str(i)] = x
            p['yS'+str(i)] = np.mean(sp[w])

    fitOnly = list(filter(lambda k: k.startswith('yS'), p.keys()))
    fitOnly.append('dl0')

    doNotFit = []
    doNotFit.extend(list(filter(lambda k: k.startswith('xS'), p.keys())))
    doNotFit.extend(['wl0'])

    # == TEST
    # plt.close(0)
    # plt.figure(0)
    # plt.plot(wl[w], sp[w], '.-k')
    # plt.plot(wl[w], Ftran(wl[w], p), '-r')
    # print(p)
    # return
    # == end TEST

    fit = {'model':Ftran(wl[w], p), 'best':p}
    if True:
        if not quiet:
            print('> fit only spectrum shape, not tellurics')
        #print('wl', wl.shape, 'sp', sp.shape)
        fit = dpfit.leastsqFit(Ftran, wl[w], p, sp[w], verbose=0, fitOnly=fitOnly, maxfev=1000)
        if MR:
            doNotFit.extend(['kern', 'kernp', 'pwv', 'pow'])
        else:
            doNotFit.extend(['kern_min', 'kern_max', 'kernp_min', 'kernp_max', 'pwv', 'pow'])

        if not quiet:
            print('> fit all but spectral resolution (kernel)')
        fit = dpfit.leastsqFit(Ftran, wl[w], fit['best'], sp[w], verbose=0,
                                doNotFit=doNotFit, maxfev=8000,
                                follow=['pwv', 'pow', 'kern', 'kern_min', 'kern_max', 'kernp'])
        if not quiet:
            print('> fit everything')
            verbose=2
        else:
            verbose=0
        if MR:
            doNotFit.remove('kern')
            doNotFit.remove('kernp')
            doNotFit.remove('pwv')
            doNotFit.remove('pow')
        else:
            doNotFit.remove('kern_min'); doNotFit.remove('kern_max')
            doNotFit.remove('kernp_min'); doNotFit.remove('kernp_max');
            doNotFit.remove('pwv')
            doNotFit.remove('pow')

        fit = dpfit.leastsqFit(Ftran, wl[w], fit['best'], sp[w], verbose=verbose,
                               doNotFit=doNotFit, maxfev=8000,
                               follow=['pwv', 'pow', 'kern', 'kern_min', 'kern_max', 'kernp', 'kernp_min', 'kernp_max'])

    p = {k:fit['best'][k] for k in fit['best'].keys() if not 'S' in k}
    if save:
        # -- SAVE telluric as fits extension in fits file
        f = fits.open(filename, mode='update')
        if 'TELLURICS' in f:
            f.pop('TELLURICS')
        c1 = fits.Column(name='EFF_WAVE', array=wl*1e-6, format='D', unit='m')
        c2 = fits.Column(name='RAW_SPEC', array=sp, format='D')
        c3 = fits.Column(name='TELL_TRANS', array=Ftran(wl, p), format='D')
        c4 = fits.Column(name='CORR_SPEC', array=sp/Ftran(wl, fit['best']), format='D')
        c5 = fits.Column(name='CORR_WAVE', array=Ftran(wl, fit['best'], retWL=True)*1e-6,
                        format='D', unit='m')
        c6 = fits.Column(name='CORR_CONT', array=Ftran(wl, fit['best'], retS=True), format='D')

        hdu = fits.BinTableHDU.from_columns([c1, c2, c3, c4, c5, c6])
        hdu.header['EXTNAME'] = 'TELLURICS'
        hdu.header['ORIGIN'] = 'https://github.com/amerand/PMOIRED'
        hdu.header['AUTHOR'] = 'amerand@eso.org'
        hdu.header['DATE'] = time.asctime()
        hdu.header['PWV'] = (round(fit['best']['pwv'], 3), 'mm of precipitable water')
        f.append(hdu)
        f.close()

    if quiet:
        return

    print('isnan:', any(np.isnan(fit['model'])))
    # -- PLOT RESULTS --------------------------------------------------
    plt.plot(wl[w], sp[w], '-c', alpha=0.5, label='raw data')
    plt.plot(wl[~w], sp[~w], '.', color='orange', alpha=0.5, label='ignored')

    #plt.plot(wl[w], fit['model'], '-k', alpha=0.2, linewidth=1, label='model')
    plt.plot(wl, Ftran(wl, fit['best']), '-k', alpha=0.2, linewidth=1,
        label='model PWV=%.2fmm'%fit['best']['pwv'])

    plt.plot(wl, sp/Ftran(wl, p)+np.median(sp)/8, '-b', label='telluric corrected + offset')
    plt.title(filename);
    plt.ylim(0.9*np.min(sp[w]), 1.1*np.max(sp[w]/Ftran(wl[w], p)+np.median(sp[w])/8))
    lines = {'HeI':[(2.0581, 2.11274, 2.11267, 2.11258, 2.161), 'c', 'dashed'],
             'HeII':((2.1885), 'c', 'dotted'),
             'MgII':((2.137, 2.143, ), '0.5', 'dashed'),
             #'MgI':((2.10655, 2.10666), '0.5', 'dotted'),
             'HI':[(2.1661, 2.360, 2.367, 2.374, 2.383, 2.392, 2.404, 2.416, 2.431, 2.449), 'b', 'dashed'],
             'H2':[(2.122), 'b', 'dotted'],
             'NaI':[(2.208, 2.208969), 'y', 'dashed'],
             'NIII':[(2.1155, 2.247, 2.251), (0,1,0.5), 'dashed'],
             'FeII':((2.089, 2.145, 2.241, 2.368), 'g', 'dashed'),
             r'$^{12}$C$^{16}$O HB':([2.2935, 2.3227, 2.3535, 2.3829, 2.4142], 'r', 'dashed'),
             r'$^{13}$C$^{16}$O HB':([2.3448, 2.3739, 2.4037, 2.4341, ], 'r', 'dotted'), #2.4971
             #'AlI':((2.109884,2.116958),(0.2,0.5,0.8), 'dotted'),
             #'MgI':((2.121393,2.146472,2.386573),'r'),
             #'ScI':((2.20581,2.20714),'m'),
             #'SiII':((2.200),'g'),
             'CaI':((1.99318, 2.261410, 2.263110, 2.265741),'g', 'dotted'), #1.89282, 1.94508, 1.98702,
             }
    for k in list(lines.keys()):
        plt.vlines(lines[k][0], plt.ylim()[0], plt.ylim()[1], color=lines[k][1],
                linestyle=lines[k][2], label=k, alpha=0.3, linewidth=1)
    plt.legend(fontsize=6, ncol=4)
    try:
        plt.tight_layout()
    except:
        pass
    return

def showTellurics(filename, fig=99):
    h = fits.open(filename)
    if not 'TELLURICS' in h:
        print('no tellurics model found')
        return
    if not fig is None:
        plt.close(fig)
        plt.figure(fig, figsize=(pmoired.FIG_MAX_WIDTH,
                                  0.4*pmoired.FIG_MAX_WIDTH))

    ax1 = plt.subplot(211)
    plt.plot(h['TELLURICS'].data['EFF_WAVE']*1e6,
             h['TELLURICS'].data['RAW_SPEC'],
             '-', alpha=0.5, label='raw spectrum', color='orange', lw=2)
    plt.plot(h['TELLURICS'].data['EFF_WAVE']*1e6,
             h['TELLURICS'].data['CORR_CONT'],
             ':g', alpha=0.5, label='estimated continuum', lw=1)
    plt.plot(h['TELLURICS'].data['EFF_WAVE']*1e6,
             h['TELLURICS'].data['RAW_SPEC']/h['TELLURICS'].data['TELL_TRANS'],
             '-k', label='corrected spectrum', lw=1)
    plt.plot(h['TELLURICS'].data['EFF_WAVE']*1e6,
             h['TELLURICS'].data['TELL_TRANS']*h['TELLURICS'].data['CORR_CONT'],
             '-b', label='telluric*continuum (PWV=%.2fmm)'%h['TELLURICS'].header['PWV'],
             alpha=0.5, lw=1)

    plt.legend()
    plt.ylabel("flux (arb. unit)")

    ax2 = plt.subplot(212, sharex=ax1)
    plt.plot(h['TELLURICS'].data['EFF_WAVE']*1e6,
         h['TELLURICS'].data['RAW_SPEC']/h['TELLURICS'].data['CORR_CONT'],
          '-', alpha=0.5, label='raw normalised spectrum', color='orange', lw=2)
    plt.plot(h['TELLURICS'].data['EFF_WAVE']*1e6,
             h['TELLURICS'].data['TELL_TRANS'],
             '-b', label='telluric model (PWV=%.2fmm)'%h['TELLURICS'].header['PWV'],
             alpha=0.5, lw=1)
    plt.plot(h['TELLURICS'].data['EFF_WAVE']*1e6,
             h['TELLURICS'].data['RAW_SPEC']/
             h['TELLURICS'].data['TELL_TRANS']/h['TELLURICS'].data['CORR_CONT'],
             '-k', label='corrected spectrum', lw=1)
    plt.legend()
    plt.ylabel("normalised flux")
    plt.xlabel(r'wavelength ($\mu$m)')
    plt.suptitle(os.path.basename(filename), fontsize=8)

    plt.tight_layout()
    h.close()
    return



from multiprocessing import Pool
import pmoired
import numpy as np
from matplotlib import pyplot as plt

def _chi2map2D(oidata, PARAM):
    return [np.mean(pmoired.oimodels.residualsOI(oidata, p)**2) for p in PARAM]

def chi2map2D(oi, expl, model=None, constrain=None, verbose=True, fig=0):
    """
    oi: pmoired OI object
    expl: {'grid':{'param1':(lower bound, upper bound, step), 'param2':(lower bound, upper bound, step)}}
    model: dict model (will try best fit by default)
    constrain: see gridFit
    """
    oi._merged = pmoired.oifits.mergeOI(oi.data, collapse=True, verbose=False)

    if model is None:
        model = oi.bestfit['best'].copy()
    # -- only for "grid"
    if not (type(expl) is dict and  'grid' in expl):
        raise Exception("'expl' should be {'grid':{}}")
    kx = sorted(expl['grid'].keys())[0]
    ky = sorted(expl['grid'].keys())[1]
    nx = int((expl['grid'][kx][1]-expl['grid'][kx][0])/expl['grid'][kx][2] + 1)
    ny = int((expl['grid'][ky][1]-expl['grid'][ky][0])/expl['grid'][ky][2] + 1)
    N = nx*ny
    if verbose:
        print(N, 'grid points')
    X = np.linspace(expl['grid'][kx][0], expl['grid'][kx][1], nx)
    Y = np.linspace(expl['grid'][ky][0], expl['grid'][ky][1], ny)
    X, Y = np.meshgrid(X, Y)
    X, Y = X.flatten(), Y.flatten()
    if constrain is None:
        constrain = []
    PARAM = []
    for i in range(len(X)):
        model[kx] = X[i]
        model[ky] = Y[i]
        res = [0]
        for p in constrain:
            form = p[0]
            val = str(p[2])
            for i in range(3):
                for k in model.keys():
                    if k in form:
                        form = form.replace(k, '('+str(model[k])+')')
                    if k in val:
                        val = val.replace(k, '('+str(model[k])+')')
            # -- residual
            resi = '0'
            if len(p)==3:
                resi = '('+form+'-'+str(val)+')'
            elif len(p)==4:
                resi = '('+form+'-'+str(val)+')/abs('+str(p[3])+')'
            if p[1]=='<' or p[1]=='<=' or p[1]=='>' or p[1]=='>=':
                resi = '%s if 0'%resi+p[1]+'%s else 0'%resi
            try:
                res.append(eval(resi))
            except:
                print('WARNING: could not compute constraint "'+resi+'"')
        if all(np.array(res)==0):
            PARAM.append(model.copy())
    if len(PARAM)<N and verbose:
        print('%d/%d'%(N-len(PARAM), N), 'grid point%s not within constraints'%('' if (N-len(PARAM))==1 else 's'))

    if True:
        pool = Pool()
        _chi2 = []
        N = 8
        for i in range(N):
            _chi2.append(pool.apply_async(_chi2map2D, (oi._merged, PARAM[i::N],)))
        pool.close()
        pool.join()
        _chi2 = [r.get(timeout=1) for r in _chi2]
        chi2 = np.zeros(len(PARAM))
        for i in range(N):
            chi2[i::N] = np.array(_chi2[i])
    else:
        # -- single thread
        chi2 = _chi2map2D(oi._merged, PARAM)

    X, Y = [p[kx] for p in PARAM], [p[ky] for p in PARAM]
    if fig:
        plt.close(fig)
        plt.figure(fig, figsize=(pmoired.FIG_MAX_HEIGHT, pmoired.FIG_MAX_HEIGHT))
        #plt.subplot(111, aspect='equal')
        X, Y = [p[kx] for p in PARAM], [p[ky] for p in PARAM]
        plt.scatter(X, Y, c=np.log10(chi2), s=10, cmap='gist_stern')
        plt.colorbar(label=r'log$_{10}$($\chi2$)')
    return X, Y, chi2



import os
import pickle
import time
import numpy as np
from matplotlib import pyplot
import multiprocessing

this_dir, this_filename = os.path.split(__file__)

directory = 'binlens_images'
directory = os.path.join(this_dir, directory)
exe = 'binlen_images_oneepoch.out'
txt = 'parameters.txt'
#txt = os.path.join(directory, txt)
assert os.path.exists(directory), 'directory "'+directory+'" not found'
if not os.path.exists(os.path.join(directory, exe)):
    if os.path.exists(os.path.join(directory, 'compile.sh')):
        print(this_filename, '> compiling', exe, end='...')
        os.system('cd '+directory+'; ./compile.sh')
        assert os.path.exists(os.path.join(directory, exe)), 'compilation failed'
        print('OK')
    else:
        assert False, 'cannot find compile script'
assert os.path.exists(os.path.join(directory, exe)), 'executable "'+exe+'" not found in "'+directory+'"'

debug = False
def clean():
    global directory, exe
    files = os.listdir(directory)
    for f in files:
        if f.startswith('img.') and (f.endswith('.pickle') or
                                     f.endswith('.txt')):
            os.system('rm '+os.path.join(directory, f))
        if f.startswith('_tmp'):
            os.system('rm -rf '+os.path.join(directory, f))
        if f in [txt]:
            os.system('rm '+os.path.join(directory, f))
    return

def computeSparseParam(MJD, param, parallel=True):
    images = computeImages(MJD, param, parallel=parallel)
    fluxes = []
    mjd = []
    for k in images:
        mjd.append(50000.5+float(os.path.basename(images[k]).split('img.')[1].split('.pickle')[0]))
        with open(images[k], 'rb') as f:
            tmp = pickle.load(f)
        fluxes.append(np.sum(tmp['I']))

    fluxes = np.array(fluxes)
    fluxes = fluxes[np.argsort(mjd)]
    mjd = sorted(mjd)
    s = 'np.interp($MJD, %s, %s)'%('['+','.join(['%.2f'%x for x in mjd])+']',
                                   '['+','.join(['%.3f'%x for x in fluxes])+']')
    return {'sparse':images, 'spectrum(mjd)':s}

def computeImages(MJD, param, parallel=False, debug=debug, directory=directory):
    global txt, exe
    if parallel:
        preptime = 0
        t0 = time.time()
        nproc = min(multiprocessing.cpu_count(), len(MJD))
        dirs = []
        for i in range(nproc):
            dirs.append(os.path.join(directory, '_tmp%03d'%i))
            os.mkdir(dirs[-1])
            os.system('cp '+os.path.join(directory, exe)+' '+os.path.join(dirs[-1], exe))

        Pool = multiprocessing.Pool()
        res = []
        for i,mjd in enumerate(MJD):
            res.append(Pool.apply_async(computeImages, ([mjd], param, ),
                {'parallel':False, 'debug':False,
                'directory':dirs[i%len(dirs)]}))
        preptime += time.time()-t0
        Pool.close()
        Pool.join()
        t1 = time.time()
        tmp = [r.get(timeout=1) for r in res]
        res = {}
        for t in tmp:
            res.update(t)
        for d in dirs:
            os.system('mv '+os.path.join(d, '*.pickle')+' '+directory)
            os.system('rm -rf '+d)
        for k in res:
            for d in dirs:
                if os.path.basename(d) in res[k]:
                    res[k] = res[k].replace(os.path.basename(d), '')
                    res[k] = res[k].replace('//', '/')
        preptime += time.time()-t1
        print('preptime: %.2fs'%preptime)
        return res

    cmd = 'cd '+directory+'; ./'+exe
    if len(MJD)==1:
        print(directory, MJD)
    t0 = time.time()
    cols = ['t0', 'u0', 'tE', 'b', 'q', 'theta', 'rhos']
    if not 't0' in param and 'mjd0' in param:
        param['t0'] = param['mjd0']-50000.5
    res = {}
    if debug:
        print('D> -- computeImage starts')
    for mjd in MJD:
        t = round(mjd -50000.5, 2)
        img = 'img.%.2f.txt'%t
        if debug:
            print('D> computing for epoch', img)
        t1 = time.time()
        with open(os.path.join(directory, txt), 'w') as f:
            tmp = ' '.join(['%f'%param[k] for k in cols])
            tmp += ' %.2f'%t
            f.write(tmp+'\n')
            if debug:
                print('D>', txt, '=', '"'+tmp+'"')
        if debug:
            print('D> executing "'+cmd+'"')
        os.system(cmd)
        if debug:
            print('D> image computed in %.1fs'%(time.time()-t1))
        t1 = time.time()
        imageTxt2pickle(os.path.join(directory, img),
                        rhos=param['rhos'], debug=debug)
        if debug:
            print('D> conversion in %.1fs'%(time.time()-t1))
        os.system('rm '+os.path.join(directory, img))
        res[round(mjd, 2)] = os.path.join(directory, img.replace('.txt', '.pickle'))
    if debug:
        print('D> -- computeImage done in %.1fs'%(time.time()-t0))
    return res

def imageTxt2pickle(filename, debug=debug, rhos=None):
    tinit = time.time()
    data = []
    with open(filename) as f:
        for l in f.readlines():
            data.append((np.double(l.split()[0]), np.double(l.split()[1])))
    X, Y = np.array([d[0] for d in data]), np.array([d[1] for d in data])
    x0 = np.mean(X)
    y0 = np.mean(Y)
    dx = 0.005 # original file

    dx = 0.005*4 # binned

    print('X:', min(X), max(X))
    print('Y:', min(Y), max(Y))


    nx, ny = int(X.ptp()/dx)+1, int(Y.ptp()/dx)+1

    x = np.linspace(X.min(), X.max(), nx)
    y = np.linspace(Y.min(), Y.max(), ny)

    _X, _Y = np.meshgrid(x, y)
    _I = np.zeros(_X.shape)
    t = time.time()
    _i = np.int_([(d[1]-y.min())/dx for d in data])
    _j = np.int_([(d[0]-x.min())/dx for d in data])
    for k,d in enumerate(data):
        _I[_i[k], _j[k]] += 1

    # -- generate sparse image
    sparse = {'x':[], 'y':[], 'I':[]}
    for i in range(ny):
        for j in range(nx):
            if _I[i,j]>0:
                sparse['x'].append(_X[i,j])
                sparse['y'].append(_Y[i,j])
                sparse['I'].append(_I[i,j])
    for k in sparse:
        sparse[k] = np.array(sparse[k])

    if not rhos is None:
        # -- normalise fluxes to flux of source
        # -- not sure how to justify the "0.002"!
        fscale = (0.002/rhos)**2/np.pi
        sparse['I'] *= fscale
    fileb = filename.replace('.txt', '.pickle')
    if debug:
        print('D> saving in', fileb)
    with open(fileb, 'wb') as f:
        pickle.dump(sparse, f)
    return fileb
